/*

NK Visualization

Copyright (c) 2010 Erkin Bahceci <erkin@cs.utexas.edu>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

*/

#include "GUI.h"

#include <QMenu>

#include "vtkRenderWindow.h"
#include "vtkRenderer.h"
#include "vtkCommand.h"
#include "vtkEventQtSlotConnect.h"

GUI::GUI (Simulation &simulation,
          VisualizationType visType,
          long startPoint) :
    mSimulation (simulation),
    mVisualization (NULL),
    mTerrainVisualization (NULL),
    mSphericalPolarVisualization (NULL),
    mDotVisualization (NULL),
    mVisType (visType)
{
    this->setupUi(this);

    switch (mVisType)
    {
    case VisTypeTerrain:
	mTerrainVisualization =
	    new TerrainVisualization (simulation);
	mVisualization = dynamic_cast<Visualization*> (mTerrainVisualization);
	mDotVisualization =
	    dynamic_cast<DotVisualization*> (mTerrainVisualization);
	break;
#if 0
    case VisTypeTerrainAnim:
	mAnimTerrainVisualization =
	    new AnimTerrainVisualization (simulation, labelCoords);
	mVisualization =
	    dynamic_cast<Visualization*> (mAnimTerrainVisualization);
	mDotVisualization =
	    dynamic_cast<DotVisualization*> (mAnimTerrainVisualization);
	break;
#endif
    case VisTypeSphericalPolar:
	mSphericalPolarVisualization =
	    new SphericalPolarVisualization (simulation, startPoint);
	mVisualization =
	    dynamic_cast<Visualization*> (mSphericalPolarVisualization);
	mDotVisualization =
	    dynamic_cast<DotVisualization*> (mSphericalPolarVisualization);
	break;
    default:
	break;
    }

    mVisualization->init ();

    // give the render window to QVTKWidget
    vtkRenderWindow* renWin = mVisualization->getRenderWindow ();
    qVTKMain->SetRenderWindow (renWin);

#if 0
    if (visType == VisTypeTerrainAnim)
    {
	// Construct a 1-second timeline with a frame range of 0 - 100
	mTimeLine1 = new QTimeLine (1000, this);
	//mTimeLine1->setFrameRange (0, 1000);
	mTimeLine1->setCurveShape (QTimeLine::LinearCurve);
	mTimeLine1->setUpdateInterval (16);
	connect (mTimeLine1, SIGNAL(valueChanged(qreal)), this, SLOT(animTick1(qreal)));
	connect (mTimeLine1, SIGNAL(finished()), this, SLOT(animFinish1()));

	// Clicking the button will start the animation
	connect (buttonNext, SIGNAL(clicked()), mTimeLine1, SLOT(start()));
    }
    else
#endif
	initCallbacks ();
}

GUI::~GUI()
{
}

void GUI::initCallbacks ()
{
    if (mVisType == VisTypeTerrain)
    {
	connect (checkBoxGlobalMax, SIGNAL (toggled (bool)), this,
		 SLOT (toggleGlobalMax (bool)));
	connect (checkBoxLocalMax, SIGNAL (toggled (bool)), this,
		 SLOT (toggleLocalMax (bool)));
	connect (checkBoxNeighbors, SIGNAL (toggled (bool)), this,
		 SLOT (toggleNeighborArcs (bool)));

	// Hide irrelevant controls
	buttonCenter->setVisible (false);
	lineEditCenter->setVisible (false);
	labelOrigin->setVisible (false);
    }
    else if (mVisType == VisTypeSphericalPolar)
    {
	connect (buttonCenter, SIGNAL (clicked ()), this,
		 SLOT (centerSelectedPoint ()));

	connect (lineEditCenter, SIGNAL (editingFinished ()), this,
		 SLOT (handleNorthPoleTextEdit ()));

	// Update edit box
	QString str;
	str.sprintf("%ld", mSphericalPolarVisualization->northPolePoint ());
	lineEditCenter->setText(str);

	// Hide irrelevant controls
	checkBoxNeighbors->setVisible (false);
	checkBoxGlobalMax->setVisible (false);
	checkBoxLocalMax->setVisible (false);
	groupBoxShow->setVisible (false);
    }
}

void GUI::postSteps ()
{
}

void GUI::centerSelectedPoint ()
{
    mSphericalPolarVisualization->centerSelectedPoint ();
    updateAfterNorthPoleChange ();
}

void GUI::handleNorthPoleTextEdit ()
{
    long newPole = lineEditCenter->text ().toLong ();
    if (newPole != mSphericalPolarVisualization->northPolePoint ())
    {
	mSphericalPolarVisualization->centerPoint (newPole, true);
	updateAfterNorthPoleChange ();
    }
}

void GUI::updateAfterNorthPoleChange ()
{
    // Update edit box
    QString str;
    str.sprintf("%ld", mSphericalPolarVisualization->northPolePoint ());
    lineEditCenter->setText(str);

    qVTKMain->update();
}

void GUI::updateAll ()
{
    mVisualization->update ();
    qVTKMain->update();
}

void GUI::toggleNeighborArcs (bool newState)
{
    mTerrainVisualization->toggleNeighborArcs (newState);
    qVTKMain->update();
}

void GUI::toggleGlobalMax (bool newState)
{
    mDotVisualization->toggleGlobalMax (newState);
    qVTKMain->update();
}

void GUI::toggleLocalMax (bool newState)
{
    mDotVisualization->toggleLocalMax (newState);
    qVTKMain->update();
}
#if 0
void GUI::animTick1 (double progress)
{
    mAnimTerrainVisualization->tick1(progress);
}

void GUI::animFinish1 ()
{
    mAnimTerrainVisualization->finish1();
    initCallbacks ();
}
#endif
