/*

NK Visualization

Copyright (c) 2010 Erkin Bahceci <erkin@cs.utexas.edu>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

*/

#include <qapplication.h>
#include "GUI.h"

#include "nk.h"
#include "simulation-io.h"

// ************* Simulation *****************

Simulation::Simulation (const string &simuCfgFileName) :
    params (simuCfgFileName),
    landscape (params)
{
}

Simulation::~Simulation ()
{
}

/// Do one visualized run
void
Simulation::executeVisualized (int argc, char *argv[],
			       VisualizationType visType,
			       long startingPoint)
{
    params.print ();

    initRun ();

    QApplication app (argc, argv);

    GUI widget (*this, visType, startingPoint);
    widget.show ();
    widget.move (680 - widget.x (), 200 - widget.y ());

    app.exec ();

    finishRun ();
}

/// Initialize a run
void
Simulation::initRun ()
{
    landscape.initRun (0);
}

/// Finalize a run
void
Simulation::finishRun ()
{
    landscape.finishRun ();
}

void Simulation::initRandomState(unsigned short *randomState,
                                 unsigned long seed)
{
    unsigned long mask = ~0L >> 48;
    randomState[0] = (unsigned short)(seed & mask);
    randomState[1] = (unsigned short)((seed & (mask << 16)) >> 16);
    randomState[2] = (unsigned short)((seed & (mask << 32)) >> 32);
}

// ************* Simulation Parameters *****************

SimuParameters::SimuParameters (const string &simuCfgFileName) :
    n (10),
    k (4),
    nPoints (1024),
    nContexts (16),
    landscapeseed (0)
{
    if (simuCfgFileName.length() > 0)
	readFromFile (simuCfgFileName);
}

SimuParameters::~SimuParameters ()
{
}

void
SimuParameters::readFromFile (const string &simuCfgFileName)
{
    FILE *fp = 0;                // pointer to the simulation config. file

    open_file (simuCfgFileName.c_str (), "r", &fp, true);

    // get n and k
    read_till_keyword (fp, PARAM_N, true);
    fscanf (fp, "%u", &n);
    read_till_keyword (fp, PARAM_K, true);
    fscanf (fp, "%u", &k);
    if (n < 1 || n > MAXN || k > n-1)
    {
	errorExit ("n=%u or k=%u out of bounds\n", n, k);
    }

    nPoints = 1L << n;
    nContexts = 1L << k;

    // get seeds; seed 0 makes the first generated number 0;
    // seeds are mod 65536 since they are cast unsigned short

    read_till_keyword (fp, PARAM_LANDSCAPE_SEED, true);
    fscanf (fp, "%ld", &landscapeseed);

    fclose (fp);
}

/// Prints out the read and computed values.
void
SimuParameters::print ()
{
    printf ("n %d\nk %d\n", n, k);

    printf ("landscapeseed %ld\n", landscapeseed);
    printf ("\nnpoints %lu\nncontexts %lu\n", nPoints, nContexts);
}
