/*

NK Visualization

Copyright (c) 2010 Erkin Bahceci <erkin@cs.utexas.edu>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

*/

#include "nk.h"

#define MAXSTRL 1000            /* max length of input lines (chars) */

// ************* Point *****************

unsigned Point::n = 0;
unsigned Point::kPlus1Consecutive = 0;
long Point::rightKPlus1Filter = 0;
long Point::rightKFilter = 0;

/// create a new point with the value at the given dimension flipped
Point::Point (const Point &point, unsigned dim) :
    values (point.values)
{
    values ^= (1L << dim);
}

/// Create a new point from a given point by flipping D of its dimensions,
/// where D is in the range [minJumpDist, maxJumpDist].
Point::Point (const Point &point, int minJumpDist, int maxJumpDist,
	      unsigned short *searchRandomState) :
    values (point.values)
{
    long changeMask = 0;
    int dims2Change = minJumpDist +
	(int) (erand48 (searchRandomState) * (maxJumpDist + 1 - minJumpDist));

    // If all dim.s are to be changed, just flip them all
    if (dims2Change == (int)n)
    {
	values ^= (1L << n) - 1;
	return;
    }

    for (int changedDims = 0; changedDims < dims2Change; changedDims++)
    {
	// Pick the unchanged-bit count for the position of the dim. to flip
	int unchangedDimCount =
	    1 + (int) (erand48 (searchRandomState) * ((int)n - changedDims));

	long bitMask = 1;
	for (int i = 0; ; bitMask <<= 1)
	{
	    if (changeMask & bitMask) // dim. already changed
		continue;

	    // Found unchanged dim.
	    i++;
	    if (i == unchangedDimCount)
		break;
	}

	changeMask |= bitMask; // bitMask here indicates the bit to change
    }

    values ^= changeMask; // Flip the bits indicated by mask
}

void
Point::print (bool goToNextLine) const
{
    printf ("%4ld", values);
    if (goToNextLine)
	printf ("\n");
}

string
Point::toBitString () const
{
    string s = "";
    long v = values;
    for (unsigned dim = 0; dim < n; dim++, v >>= 1)
	s = string ((v & 1) ? "1" : "0") + s;
    return s;
}

long
Point::getHash () const
{
    return values;
}

/// get the bit value of the given dimension
unsigned
Point::getValue (unsigned dim) const
{
    return (values >> dim) & 1;
}

/// Calculate the hamming distance between this point and another one
int
Point::hammingDist(long v) const
{
    v ^= values;  // XOR the two values to find their difference
    v -= ((v >> 1) & 0x55555555);  // reuse input as temporary
    v = (v & 0x33333333) + ((v >> 2) & 0x33333333);  // temp
    return (((v + (v >> 4)) & 0xF0F0F0F) * 0x1010101) >> 24;  // count
}

Point &
Point::operator= (const Point &otherPoint)
{
    values = otherPoint.values;
    return *this;
}

bool
Point::operator== (const Point &otherPoint) const
{
    return (values == otherPoint.values);
}

bool
Point::operator!= (const Point &otherPoint) const
{
    return (values != otherPoint.values);
}


// ************* input and output *****************

/** open a file, exit or return false if failure */
bool
open_file (const char *filename,// file name
	   const char *mode,    // return the file pointer
	   FILE **fp,           // open mode: "r", "a", "w"
	   bool required)        // whether to exit if cannot open
{
    if ((*fp = fopen (filename, mode)) == NULL)
    {
	fprintf (stderr, "Cannot open %s\n", filename);
	if (required)
	{
	    errorExit ("");
	}
	else
	    return false;
    }
    return true;
}

/// get rid of the rest of the line and the newline in the end
void
fgl (FILE *fp)
{
    int c;

    while ((c = getc (fp)) != EOF && c != '\n')
	;
}

/** read text in the file until you find the given keyword
   in the beginning of the line; if the keyword is required but
   not found, terminate program; otherwise return if it was found or not */
bool
read_till_keyword (FILE * fp,       // pointer to the file
		   const char keyword[],  // the keyword string
		   bool required)    // whether to abort if keyword not found
{
    char s[MAXSTRL + 1];
    long pos = ftell (fp);

    while (true)
    {
	if (fscanf (fp, "%s", s) == EOF)
	{
	    /* file ran out, keyword was not found */
	    if (required)
	    {
		fprintf (stderr, "Could not find keyword: %s\n", keyword);
		errorExit ("");
	    }
	    else
	    {
	    	/* that's ok; rewind the file back to where we started and
	    	   return the signal indicating that it was not found */
	    	fseek (fp, pos, SEEK_SET);
		return false;
	    }
	}
	else
	/* something was read from the file */
	{
	    /* if it was the keyword, return success signal */
	    if (!strcasecmp (s, keyword))
		return true;
	    /* otherwise get rid of the rest of the line */
	    fgl (fp);
	}
    }
}

