/*

NK Visualization

Copyright (c) 2010 Erkin Bahceci <erkin@cs.utexas.edu>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

*/

#include "visualization-dot.h"

#include <vtkCamera.h>
#include <vtkCellArray.h>
#include <vtkFloatArray.h>
#include <vtkPointData.h>
#include <vtkPolyData.h>
#include <vtkPolyDataMapper.h>
#include <vtkPolyDataMapper2D.h>
#include <vtkProperty.h>
#include <vtkProperty2D.h>
#include <vtkTextProperty.h>
#include <vtkObjectFactory.h>


double DotVisualization::dotColors[DotTypeNum][4] =
    {{1, 0, 1, 0.5},
     {0, 1, 1, 0.5}};

vtkCxxRevisionMacro(CustomLegendBoxActor, "$Revision: 1.0 $");
vtkStandardNewMacro(CustomLegendBoxActor);

//======= DotVisualization =======

DotVisualization::DotVisualization (Simulation &simulation,
				    VisualizationType visType,
				    QLabel *labelPoint) :
    Visualization (simulation, visType, labelPoint),
    mNumDotGroups (0),
    mGlobalMaxOn (false),
    mLocalMaxOn (false)
{
    for (int i = 0; i < DotTypeNum; i++)
	mDotGroups[i] = NULL;

    mLegendActor = CustomLegendBoxActor::New ();
}

DotVisualization::~DotVisualization ()
{
    removeAllDotGroups ();
    mLegendActor->Delete ();
}

void
DotVisualization::removeAllDotGroups ()
{
    for (int i = 0; i < mNumDotGroups; i++)
	if (mDotGroups[i])
	{
	    delete mDotGroups[i];
	    mDotGroups[i] = NULL;
	}
    mNumDotGroups = 0;
}

void
DotVisualization::toggleGlobalMax (bool newState)
{
    mGlobalMaxOn = newState;

    // Update dots
    updatePointDots ();
}

void
DotVisualization::toggleLocalMax (bool newState)
{
    mLocalMaxOn = newState;

    // Update dots
    updatePointDots ();
}

/// Shows individual memory points with dots
void
DotVisualization::updatePointDots ()
{
    createLegendBox ();

    if (mNumDotGroups)
	removeAllDotGroups ();

    if (mVisType == VisTypeTerrain)
    {
	if (mLocalMaxOn)
	{
	    int numLocalMax = mLocalMaxPoints.size (); // TODO
	    vector<double> pointCoords;
	    pointCoords.reserve (3 * numLocalMax);

	    foreach (const Point &localMaxPoint, mLocalMaxPoints)
		insertPointDot (localMaxPoint, DotTypeLocalMax, &pointCoords);

	    mDotGroups[mNumDotGroups++] =
		new DotGroup (mRenderer, numLocalMax, &pointCoords[0],
			      dotColors[DotTypeLocalMax]);
	}
	if (mGlobalMaxOn)
	{
	    vector<double> pointCoords;
	    pointCoords.reserve (3);

	    insertPointDot (mGlobalMaxPoint, DotTypeGlobalMax, &pointCoords);

	    mDotGroups[mNumDotGroups++] =
		new DotGroup (mRenderer, 1, &pointCoords[0],
			      dotColors[DotTypeGlobalMax]);
	}
    }

    postUpdateDots ();
}

void
DotVisualization::addDotsForPointList (const vector<Point> &pointList,
                                       int start,
                                       int numPoints,
				       DotType dotType,
				       vector<double> &pointCoords)
{
    // Optimize push_backs by reserving here enough space to hold all points
    pointCoords.reserve (3 * numPoints);

    int end = start + numPoints;
    for (int i = start; i < end; i++)
	insertPointDot (pointList[i], dotType, &pointCoords);
}

void
DotVisualization::createLegendBox ()
{
    if (mVisType != VisTypeTerrain)
	return; // only for rectangular terrain visualization for now

    int nEntries = 2;
    const char *labels[] = {"Global max",
			    "Local max"};

    mLegendActor->SetNumberOfEntries (nEntries);
    mLegendActor->SetPadding (8);
    mLegendActor->SetPosition (0.8, 0.9);
    mLegendActor->SetPosition2 (0.199, 0.08);

    mLegendActor->GetProperty ()->SetColor (0, 0, 0);
    mLegendActor->GetProperty ()->SetPointSize (10);

    vtkLookupTable *colorTable = vtkLookupTable::New ();
    colorTable->SetNumberOfColors (nEntries);
    for (int i = 0; i < nEntries; i++)
	colorTable->SetTableValue (i, dotColors[i]);
    colorTable->Build ();

    for (int i = 0; i < nEntries; i++)
    {
	vtkPoints* pts = vtkPoints::New();
	vtkCellArray* conn = vtkCellArray::New();
	vtkPolyData* poly = vtkPolyData::New();

	vtkIdType pntId = pts->InsertNextPoint (0, 18, 0);
	conn->InsertNextCell (1, &pntId);
	poly->SetPoints (pts);
	poly->SetVerts (conn);

	// Color as scalar
	vtkFloatArray *pointsFitness = vtkFloatArray::New();
	pointsFitness->InsertNextTuple1 ((double)i / (nEntries - 1));
	poly->GetPointData ()->SetScalars (pointsFitness);

	mLegendActor->SetEntryString (i, labels[i]);
	mLegendActor->SetEntrySymbol (i, poly);
	mLegendActor->GetSymbolMapper (i)->SetLookupTable (colorTable);
    }

    vtkTextProperty *tProp = vtkTextProperty::New ();
    tProp->SetColor (0, 0, 0);
    mLegendActor->SetEntryTextProperty (tProp);

    mRenderer->AddActor (mLegendActor);
}


//======= DotGroup =======

DotGroup::DotGroup (vtkRenderer *renderer, int numDots, double *coords,
		    double colorRed, double colorGreen, double colorBlue,
		    double colorAlpha) :
    mRenderer (renderer),
    mActor (vtkActor::New ()),
    mPoints (vtkPoints::New())
{
    init (numDots, coords, colorRed, colorGreen, colorBlue, colorAlpha);
}

DotGroup::DotGroup (vtkRenderer *renderer, int numDots, double *coords,
		    double *colorRGBA) :
    mRenderer (renderer),
    mActor (vtkActor::New ()),
    mPoints (vtkPoints::New())
{
    init (numDots, coords,
          colorRGBA[0], colorRGBA[1], colorRGBA[2], colorRGBA[3]);
}

DotGroup::~DotGroup ()
{
    mRenderer->RemoveActor (mActor);
    mActor->Delete ();
    mPoints->Delete ();
}

void
DotGroup::init (int numDots, double *coords,
                double colorRed, double colorGreen, double colorBlue,
                double colorAlpha)
{
    vtkCellArray* conn = vtkCellArray::New();
    vtkPolyData* poly = vtkPolyData::New();

    for (int i = 0; i < numDots; i++, coords += 3)
    {
	vtkIdType pntId =
	    mPoints->InsertNextPoint (coords[0], coords[1], coords[2]);
	conn->InsertNextCell (1, &pntId);
    }
    poly->SetPoints (mPoints);
    poly->SetVerts (conn);

    vtkPolyDataMapper* polyMapper = vtkPolyDataMapper::New();
    polyMapper->SetInput(poly);

    mActor->SetMapper(polyMapper);
    mActor->GetProperty()->SetColor (colorRed, colorGreen, colorBlue);
    mActor->GetProperty()->SetOpacity (colorAlpha);
    update (true);

    mRenderer->AddActor(mActor);
}

/// Updates all points' common size using first point's distance to the camera,
/// so that they look more realistic.
void
DotGroup::update (bool spherical, int numDims)
{
    vtkCamera *camera = mRenderer->GetActiveCamera();
    if (!camera)
	return;

    double *camPos = camera->GetPosition ();
    double centerPos1[3] = {0, 0, 0};
    double centerPos2[3] = {0.5, 0.5, 0};
    double *centerPos;

    if (spherical)
    {
	centerPos = centerPos1;
    }
    else
	centerPos = centerPos2;

    double distance = sqrt ((camPos[0] - centerPos[0]) *
			   (camPos[0] - centerPos[0]) +
			   (camPos[1] - centerPos[1]) *
			   (camPos[1] - centerPos[1]) +
			   (camPos[2] - centerPos[2]) *
			   (camPos[2] - centerPos[2]));
    double sizeFactor = 15 * 12.0 / numDims;
    double size = sizeFactor / distance;
    mActor->GetProperty()->SetPointSize (size);
}
