/*

NK Visualization

Copyright (c) 2010 Erkin Bahceci <erkin@cs.utexas.edu>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

*/

#include "visualization-polar.h"

#include <vtkActor.h>
#include <vtkCamera.h>
#include <vtkProperty.h>
#include <vtkProperty2D.h>
#include <vtkTextProperty.h>
#include <vtkTubeFilter.h>
#include <vtkPolyDataMapper.h>
#include <vtkLookupTable.h>
#include <vtkFloatArray.h>
#include <vtkCellArray.h>
#include <vtkPointData.h>
#include <vtkPoints.h>
#include <vtkPolyDataMapper.h>
#include <vtkExtractEdges.h>
#include <vtkCleanPolyData.h>
#include <vtkQuad.h>
#include <vtkUnstructuredGridReader.h>
#include <vtkDataSetSurfaceFilter.h>
#include <vtkUnstructuredGridGeometryFilter.h>
#include <vtkLinearExtrusionFilter.h>
#include <vtkPointPicker.h>
#include <vtkArrowSource.h>
#include <vtkGlyph3D.h>


// ============  Polar Visualization  ==============


/*
  Since this method generates bit strings that have at most
  2 alternating groups in a ring placement, it leaves out all strings
  that have 4 or more alternating groups in a ring placement.

n=4:
  -0123 -> 0000,0001,0011,0111,1111
  -1230 ->     ,0010,0110,1110,
  -2301 ->     ,0100,1100,1101,
  -3012 ->     ,1000,1001,1011,

  2 + n*(n-1) = 14 -> left out 2: 0101,1010

n=5:
  -01234 -> 00000,00001,00011,00111,01111,11111
  -12340 ->      ,00010,00110,01110,11110,
  -23401 ->      ,00100,01100,11100,11101,
  -34012 ->      ,01000,11000,11001,11011,
  -40123 ->      ,10000,10001,10011,10111,

  2 + n*(n-1) = 22 -> left out 10:
       00101,01010,10100,01001,10010,
       11010,10101,01011,10110,01101
*/


// ============  SphericalPolarVisualization  ==============

SphericalPolarVisualization::SphericalPolarVisualization
	(Simulation &simulation,
	 long startingPoint,
	 VisualizationType visType) :
    Visualization (simulation, visType),
    DotVisualization (simulation, visType),
    mNorthPolePoint (startingPoint),
    mNorthTheta (0),
    mNorthPhi (0),
    mSelectedTheta (0),
    mSelectedPhi (0),
    mSelectedPntInd (0),
    mSelectedPointGroup (NULL),
    mBaseRadius (0.2)
{
    mGridLinesActor = vtkActor::New ();
    mTerrainActor = vtkActor::New();

    mCaptionActor = vtkCaptionActor2D::New ();

    for (int i = 0; i < 3; i++)
	mPrevZeroPos[i] = 0;
}

SphericalPolarVisualization::~SphericalPolarVisualization ()
{
    mGridLinesActor->Delete ();
    mTerrainActor->Delete ();
    mTerrainGrid->Delete ();

    if (mSelectedPointGroup)
	delete mSelectedPointGroup;
    mCaptionActor->Delete ();
}

void
SphericalPolarVisualization::init ()
{
    Visualization::init ();

    // Set up the landscape grid
    mNumGridPoints = mNumDims * (mNumDims - 1) + 2; // 2 points for poles
    mSelectedPntInd = mNumGridPoints - 2;

    mPoints = vtkPoints::New ();
    mPoints->Allocate (mNumGridPoints);

    mScaleFactor = 1.0 / mNumDims;
    mScaleZFactor = 0.03; //3.0 / (2 * mNumDims);

    createTerrainGrid ();

    mRenderer->SetBackground (0.9, 0.9, 0.9);

    mRenWin->PointSmoothingOn (); // Makes dots circular
    mRenderer->ResetCamera();
    vtkCamera *camera = mRenderer->GetActiveCamera();
    camera->Elevation(-80.0);
    camera->OrthogonalizeViewUp();
    camera->Azimuth(135.0);
}

void
SphericalPolarVisualization::createTerrainGrid (bool update, bool reset)
{
    if (update)
    {
	mRenderer->RemoveActor(mTerrainActor);
	mRenderer->RemoveActor(mGridLinesActor);
	mTerrainGrid->Delete ();

	mGridLinesActor->Delete ();
	mGridLinesActor = vtkActor::New ();
    }

    //VTK_CREATE_NEW (vtkPoints, points);
    //mPointData =
    //static_cast<vtkFloatArray *> (points->GetData ())->GetPointer (0);

    VTK_CREATE_NEW (vtkFloatArray, pointsFitness);
    pointsFitness->Allocate (mNumGridPoints);
    mPointFitnessData = pointsFitness->GetPointer (0);

    mTerrainGrid = vtkUnstructuredGrid::New ();

    computeTerrain (!update || reset);

    mTerrainGrid->SetPoints (mPoints);
    mTerrainGrid->GetPointData()->SetScalars(pointsFitness);

    // Create grid lines

    VTK_CREATE_NEW (vtkExtractEdges, gridLines);
    gridLines->SetInput (mTerrainGrid);

    // remove duplicates to avoid the "vtkTubeFilter can't create normals" error
    VTK_CREATE_NEW (vtkCleanPolyData, cleanFilter);
    cleanFilter->SetInput (gridLines->GetOutput ());
    cleanFilter->Update ();

    VTK_CREATE_NEW (vtkTubeFilter, tubes);
    tubes->SetInputConnection (cleanFilter->GetOutputPort ());
    tubes->SetRadius (0.00075 * 10 / mNumDims);
    tubes->SetVaryRadiusToVaryRadiusOff();
    tubes->SetNumberOfSides(6);

    VTK_CREATE_NEW (vtkPolyDataMapper, gridLinesMapper);
    gridLinesMapper->SetInputConnection (tubes->GetOutputPort ());
    gridLinesMapper->SetScalarRange(0, 1);
    gridLinesMapper->ScalarVisibilityOff ();

    mGridLinesActor->SetMapper(gridLinesMapper);
    mGridLinesActor->GetProperty ()->SetColor (0, 0.3, 1);
    mGridLinesActor->GetProperty ()->SetOpacity (0.2);

    mRenderer->AddActor(mGridLinesActor);


    // Create the filter
    VTK_CREATE_NEW (vtkUnstructuredGridGeometryFilter, geom);
    geom->MergingOff ();
    geom->SetInput (mTerrainGrid);
    geom->Update (); //So that we can call GetRange() on the scalars

    mDataSetSurfaceFilter = vtkSmartPointer<vtkDataSetSurfaceFilter>::New ();
    mDataSetSurfaceFilter->PassThroughPointIdsOn ();
    mDataSetSurfaceFilter->PassThroughCellIdsOn ();
    mDataSetSurfaceFilter->SetInputConnection (geom->GetOutputPort());
    mDataSetSurfaceFilter->Update ();

    VTK_CREATE_NEW (vtkPolyDataMapper, terrainMapper);
    terrainMapper->SetInputConnection (mDataSetSurfaceFilter->GetOutputPort());
    terrainMapper->SetScalarRange(0, 1);
    terrainMapper->ScalarVisibilityOn ();
    terrainMapper->SetLookupTable(mColorTable);

    mTerrainActor->SetMapper(terrainMapper);
    mTerrainActor->GetProperty ()->LightingOff ();

    mRenderer->AddActor(mTerrainActor);
}

/// Rotate in ix-iy plane, keeping iz constant
static void
rotate2D (double src[3], double dest[3], double angle,
          unsigned ix, unsigned iy, unsigned iz)
{
    dest[ix] = cos (angle) * src[ix] - sin (angle) * src[iy];
    dest[iy] = sin (angle) * src[ix] + cos (angle) * src[iy];
    dest[iz] = src[iz];
}

void
SphericalPolarVisualization::rotateByNorthPole (double pos[3], bool reversed)
{
    double t[3];
    double theta = mNorthTheta;
    double phi = -mNorthPhi;

    if (reversed)
    {
	rotate2D (pos, t, -theta, 0, 1, 2); // Rotate in x-y plane
	rotate2D (t, pos, -phi,   0, 2, 1); // Rotate in x-z plane
    }
    else
    {
	rotate2D (pos, t, phi,   0, 2, 1); // Rotate in x-z plane
	rotate2D (t, pos, theta, 0, 1, 2); // Rotate in x-y plane
    }
}

/// Calculate coordinates for a subset of the points
void
SphericalPolarVisualization::computeTerrain (bool initializing)
{
    Landscape &landscape = mSimulation.getLandscape ();

    mNDimPoints.resize (mNumGridPoints);

    // North pole
    unsigned pointInd = mNumGridPoints - 2;
    Point pnt (mNorthPolePoint);
    double fitness = landscape.getFitnessValue (pnt);
    double posNorth[3] = {0, 0, mBaseRadius + fitness * mScaleZFactor};
    rotateByNorthPole (posNorth);
    mPoints->Reset();
    mPoints->InsertPoint (pointInd, posNorth);
    mNDimPoints[pointInd] = pnt.getHash ();
    mPointFitnessData[pointInd] = fitness;

    if (initializing)
    {
	// The north pole index should become the selected point index
	mSelectedPntInd = pointInd;

	selectPoint (mSelectedPntInd, posNorth[2], 0, 0);
    }

    pointInd++;

    // South pole
    pnt = Point (((1L << mNumDims) - 1) ^ mNorthPolePoint);
    fitness = landscape.getFitnessValue (pnt);
    double posSouth[3] = {0, 0, -(mBaseRadius + fitness * mScaleZFactor)};
    rotateByNorthPole (posSouth);
    mPoints->InsertPoint (pointInd, posSouth);
    mNDimPoints[pointInd] = pnt.getHash ();
    mPointFitnessData[pointInd] = fitness;

    for (int c = 0; c < mNumDims; c++)
    {
	Point pnt (mNorthPolePoint); // start with the north pole
	for (int r = 0; r < mNumDims - 1; r++) // end right before south pole
	{
	    pnt = Point (pnt, (r + c) % mNumDims);
	    double fitness = landscape.getFitnessValue (pnt);

	    // Rectangular coords
	    double z = fitness * mScaleZFactor;
	    double y = (r + 1) * mScaleFactor;//(double)(mNumDims - 1);
	    double x = c * mScaleFactor;

	    // Spherical coords
	    double radius = mBaseRadius + z;
	    double phi = M_PI * y;
	    double theta = -(2 * M_PI * x + r * M_PI / mNumDims);

	    double pos[3];
	    pos[0] = radius * cos (theta) * sin (phi);
	    pos[1] = radius * sin (theta) * sin (phi);
	    pos[2] = radius * cos (phi);
	    rotateByNorthPole (pos);

	    pointInd = r * mNumDims + c;
	    mPoints->InsertPoint (pointInd, pos);

	    mNDimPoints[pointInd] = pnt.getHash ();
	    mPointFitnessData[pointInd] = fitness;
	}
    }

    // Add quads (lines don't work)
    for (int c = 0; c < mNumDims; c++)
    {
	for (int r = 2; r < mNumDims - 1; r++)
	{
	    unsigned rowHeadInd = r * mNumDims;
	    addQuad (rowHeadInd + c - mNumDims,
	             rowHeadInd + (c + 1) % mNumDims - 2 * mNumDims,
		     rowHeadInd + (c + 1) % mNumDims - mNumDims,
		     rowHeadInd + c);
	}

	// Quad with the north pole point
	addQuad ((c + 1) % mNumDims,
		 c + mNumDims,
		 c,
		 mNumGridPoints - 2);

	// Quad with the south pole point
	unsigned rowHeadInd = (mNumDims - 2) * mNumDims;
	addQuad (rowHeadInd + c,
	         mNumGridPoints - 1,
	         rowHeadInd + (c + 1) % mNumDims,
	         rowHeadInd + (c + 1) % mNumDims - mNumDims);
    }
}

double
SphericalPolarVisualization::getCornerDist (const Point &point,
					    int r,
					    int c,
					    double pos[3])
{
    // take care of possibly negative c
    unsigned pointInd = r * mNumDims + (c + mNumDims) % mNumDims;
    Point pnt (mNDimPoints[pointInd]);
    mPoints->GetPoint (pointInd, pos);

    return (double) point.hammingDist (pnt.getHash()) / mNumDims;
}

void
SphericalPolarVisualization::checkIfMinDistQuad (const Point &point,
						 int r1, int c1,
						 int r2, int c2,
						 int r3, int c3,
						 int r4, int c4,
						 double &minDist,
						 double minDistPos[][3])
{
    double pos[4][3];
    double totalDist = 0;

    // Once for each of the 4 corners
    totalDist += getCornerDist (point, r1, c1, pos[0]);
    totalDist += getCornerDist (point, r2, c2, pos[1]);
    totalDist += getCornerDist (point, r3, c3, pos[2]);
    totalDist += getCornerDist (point, r4, c4, pos[3]);

    if (minDist > totalDist)
    {
	// found a more suitable region
	minDist = totalDist;
	memcpy (minDistPos[0], pos[0], 4 * 3 * sizeof (double));
    }
}

#define EQUAL_DIST_BOUND 1e-4

void
SphericalPolarVisualization::checkIfMinDistTriangle (const Point &point,
                                                     int r1, int c1,
                                                     int r2, int c2,
                                                     int r3, int c3,
                                                     DotType dotType,
                                                     double &minDist,
						     vector<SphericalPolarData> &dataBestOnes)
{
    double pos[3][3];
    double totalDist;
    double cornerDist[3];

    // Once for each of the 3 corners
    cornerDist[0] = getCornerDist (point, r1, c1, pos[0]);
    cornerDist[1] = getCornerDist (point, r2, c2, pos[1]);
    cornerDist[2] = getCornerDist (point, r3, c3, pos[2]);

    totalDist = cornerDist[0] + cornerDist[1] + cornerDist[2];

    if (totalDist < minDist + EQUAL_DIST_BOUND) // found equal-distance or closer region
    {
	if (totalDist < minDist - EQUAL_DIST_BOUND) // found a closer region
	{
	    minDist = totalDist;
	    dataBestOnes.clear (); // get rid of all "best" ones
	}

	dataBestOnes.push_back (SphericalPolarData (pos, cornerDist));
    }
}

void
SphericalPolarVisualization::checkIfMinDist2Tri (const Point &point,
						 int r1, int c1,
						 int r2, int c2,
						 int r3, int c3,
						 int r4, int c4,
						 DotType dotType,
						 double &minDist,
						 vector<SphericalPolarData> &dataBestOnes)
{
    checkIfMinDistTriangle (point,
			    r1, c1,
			    r2, c2,
			    r4, c4,
			    dotType,
			    minDist, dataBestOnes);
    checkIfMinDistTriangle (point,
			    r2, c2,
			    r3, c3,
			    r4, c4,
			    dotType,
			    minDist, dataBestOnes);
}

void
SphericalPolarVisualization::insertPointDot (const Point &point,
					     DotType dotType,
					     vector<double> *chosenList)
{
    double pointPos[3];
    positionPointDot (point, dotType, pointPos);

    // Push the point a little outward
    pushPointOutward (pointPos, 0.0018, pointPos);

    for (int i = 0; i < 3; i++)
	chosenList->push_back (pointPos[i]);
}

void
SphericalPolarVisualization::positionPointDot (const Point &point, // input pnt
					       DotType dotType,
					       double pointPos[3]) // output pos
{
    double minDist = 1e10;
    double minDistPos[4][3];

    for (int c = 0; c < mNumDims; c++)
    {
	// Regions that are not adjacent to either pole
	for (int r = 2; r < mNumDims - 1; r++)
	{
	    checkIfMinDistQuad (point,
				r, c,
				r - 1, c + 1,
				r - 2, c + 1,
				r - 1, c,
				minDist, minDistPos);
	}

	// North pole regions
	checkIfMinDistQuad (point,
			    mNumDims - 1, 0,
			    0, c,
			    1, c,
			    0, c + 1,
			    minDist, minDistPos);

	// South pole regions
	checkIfMinDistQuad (point,
			    mNumDims - 2, c,
			    mNumDims - 1, 1,
			    mNumDims - 2, c + 1,
			    mNumDims - 3, c + 1,
			    minDist, minDistPos);
    }

    // Prevent point of same memory from jumping as simulation progresses
    srand ((unsigned)point.getHash () + dotType);
    rand (); // First number is always small, discard it.

    int p1 = 1;
    int p2 = 2;

    if (DOUBLE_RAND() > 0.5)
    {
	p1 = 2;
	p2 = 3;
    }

    double v1[3] = {minDistPos[p1][0] - minDistPos[0][0],
		    minDistPos[p1][1] - minDistPos[0][1],
		    minDistPos[p1][2] - minDistPos[0][2]};
    double v2[3] = {minDistPos[p2][0] - minDistPos[0][0],
		    minDistPos[p2][1] - minDistPos[0][1],
		    minDistPos[p2][2] - minDistPos[0][2]};

    double a1 = (double)rand () / INT_MAX;
    double a2 = (double)rand () / INT_MAX;

    for (int i = 0; i < 3; i++)
	pointPos[i] = minDistPos[0][i] + a1 * v1[i] + (1 - a1) * a2 * v2[i];
}

void
SphericalPolarVisualization::update (bool reset)
{
    createTerrainGrid (true, reset);

    if (!reset)
    {
	double pos[3];
	mPoints->GetPoint (mSelectedPntInd, pos);
	double radius = sqrt (pos[0] * pos[0] +
			     pos[1] * pos[1] +
			     pos[2] * pos[2]);
	selectPoint (mSelectedPntInd, radius, mSelectedTheta, mSelectedPhi);
    }
}

void
SphericalPolarVisualization::postUpdateDots ()
{
    for (int i = 0; i < mNumDotGroups; i++)
	if (mDotGroups[i])
	    mDotGroups[i]->update (true, mNumDims);
}

void
SphericalPolarVisualization::updateOnRender ()
{
    mSelectedPointGroup->update (true, mNumDims);

    postUpdateDots ();
}

void
SphericalPolarVisualization::addQuad (vtkIdType ind0, vtkIdType ind1,
                                      vtkIdType ind2, vtkIdType ind3)
{
    vtkQuad *quad = vtkQuad::New();

    quad->GetPointIds()->SetId (0, ind0);
    quad->GetPointIds()->SetId (1, ind1);
    quad->GetPointIds()->SetId (2, ind2);
    quad->GetPointIds()->SetId (3, ind3);

    mTerrainGrid->InsertNextCell (quad->GetCellType(), quad->GetPointIds());
    quad->Delete();
}

void
SphericalPolarVisualization::selectPoint (vtkIdType pntId,
                                          double radius,
                                          double theta,
                                          double phi)
{
    mSelectedPntInd = pntId;

    radius += 0.001;
    mSelectionDotCoords[0] = radius * cos (theta) * sin (phi);
    mSelectionDotCoords[1] = radius * sin (theta) * sin (phi);
    mSelectionDotCoords[2] = radius * cos (phi);

    createSelectedPointDot ();
}

/// Returns true if a point is picked.
bool
SphericalPolarVisualization::handleClick (int x, int y)
{
    VTK_CREATE_NEW (vtkPointPicker, picker);
    picker->InitializePickList ();
    picker->AddPickList (mTerrainActor);
    picker->PickFromListOn ();

    bool picked = picker->Pick (x, y, 0, mRenderer);
    if (picked)
    {
	double pos[3];
	picker->GetPickPosition (pos);

	double radius = sqrt (pos[0] * pos[0] +
	                     pos[1] * pos[1] +
	                     pos[2] * pos[2]);

	double zOverRadius = pos[2] / radius;
	zOverRadius = MIN (1.0, zOverRadius);
	zOverRadius = MAX (-1.0, zOverRadius);

	mSelectedTheta = atan2 (pos[1], pos[0]); // original angles
	mSelectedPhi = acos (zOverRadius);

	rotateByNorthPole (pos, true);

	zOverRadius = pos[2] / radius;
	zOverRadius = MIN (1.0, zOverRadius);
	zOverRadius = MAX (-1.0, zOverRadius);

	double theta = atan2 (pos[1], pos[0]); // corrected angles
	double phi = acos (zOverRadius);

	vtkIdType pntId = 0;

	const double closeThresh = 1e-2;

	if (fabs (phi) < closeThresh)
	    pntId = mNumGridPoints - 2; // North pole
	else if (fabs (phi - M_PI) < closeThresh)
	    pntId = mNumGridPoints - 1; // South pole
	else // Non-pole point
	{
	    double frow = (phi / M_PI) / mScaleFactor - 1;
	    double fcol =
		(-theta - frow * M_PI / mNumDims) /
		(2 * M_PI * mScaleFactor);

	    if (fcol < -0.5)
		fcol += mNumDims;

	    int row = 0.5 + frow;
	    int col = 0.5 + fcol;

	    pntId = row * mNumDims + col;
	}

	Point pnt (mNDimPoints[pntId]);

	selectPoint (pntId, radius, mSelectedTheta, mSelectedPhi);
    }

    return picked;
}

void
SphericalPolarVisualization::createSelectedPointDot ()
{
    if (mSelectedPointGroup)
    {
	delete mSelectedPointGroup;
	mRenderer->RemoveActor2D (mCaptionActor);
    }
    double fitness = mPointFitnessData[mSelectedPntInd];

    mSelectedPointGroup =
	new DotGroup (mRenderer, 1, mSelectionDotCoords,
	              1, 0, 1, 1.0);
	              //fitness, fitness, fitness, 1.0);

    long pointNDimCoords = mNDimPoints[mSelectedPntInd];

    Point northPolePoint (mNorthPolePoint);
    Point selectedPoint (pointNDimCoords);

    string bitStr = selectedPoint.toBitString ();
    QString str;
    str.sprintf ("point: %ld, fitness: %.3f\n%s\ndistance to origin: %d",
                 pointNDimCoords, fitness, bitStr.c_str (),
                 northPolePoint.hammingDist (selectedPoint.getHash()));
    QByteArray ba = str.toLatin1();
    const char *c_str = ba.data();

    mCaptionActor->SetCaption (c_str);
    mCaptionActor->GetProperty ()->SetColor (1, 0, 1);
    mCaptionActor->SetAttachmentPoint (mSelectionDotCoords);
    mCaptionActor->SetPosition (25, 20);
    mCaptionActor->ThreeDimensionalLeaderOff ();
    mCaptionActor->SetWidth (0.45);//0.30);
    mCaptionActor->SetHeight (0.105);//0.075);

    vtkTextProperty *tProp = mCaptionActor->GetCaptionTextProperty ();
    tProp->ItalicOff ();
    tProp->ShadowOff ();

    mRenderer->AddActor2D (mCaptionActor);

    mSelectedPointGroup->update (true, mNumDims);
}

void
SphericalPolarVisualization::centerSelectedPoint ()
{
    mNorthPolePoint = mNDimPoints[mSelectedPntInd];
    mNorthTheta = mSelectedTheta;
    mNorthPhi = mSelectedPhi;

    // Selected point index should become the north pole index
    mSelectedPntInd = mNumGridPoints - 2;

    printf ("Resetting north pole as %ld\n", mNorthPolePoint);

    update (false);
}

void
SphericalPolarVisualization::centerPoint (long newNorthPolePoint,
					  bool resetOrientation)
{
    mNorthPolePoint = newNorthPolePoint;
    mNorthTheta = 0;
    mNorthPhi = 0;

    printf ("Resetting north pole as %ld\n", mNorthPolePoint);

    update (resetOrientation);
}

void
SphericalPolarVisualization::pushPointOutward (double inputPointPos[3],
					       double amount,
					       double outputPointPos[3])
{
    double norm = 0;
    for (int i = 0; i < 3; i++)
	norm += inputPointPos[i] * inputPointPos[i];
    norm = sqrt (norm);

    for (int i = 0; i < 3; i++)
	outputPointPos[i] = inputPointPos[i] * (1 + amount / norm);
}
