/*

NK Visualization

Copyright (c) 2010 Erkin Bahceci <erkin@cs.utexas.edu>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

*/

#include "visualization.h"

#include <vtkActor.h>
#include <vtkCamera.h>
#include <vtkCellArray.h>
#include <vtkFloatArray.h>
#include <vtkPointData.h>
#include <vtkPoints.h>
#include <vtkPolyDataMapper.h>
#include <vtkPolyDataMapper2D.h>
#include <vtkInteractorStyleTrackballCamera.h>
#include <vtkGenericRenderWindowInteractor.h>

#include <vtkProperty.h>
#include <vtkProperty2D.h>
#include <vtkTubeFilter.h>
#include <vtkLookupTable.h>
#include <vtkLight.h>
#include <vtkCommand.h>
#include <vtkParametricSpline.h>
#include <vtkParametricFunctionSource.h>
#include <vtkExtractEdges.h>
#include <vtkCleanPolyData.h>
#include <vtkPointPicker.h>
#include <vtkVectorText.h>
#include <vtkTextProperty.h>
#include <vtkCubeAxesActor.h>

#define NUM_COLORS 256

class VisRenderCallback : public vtkCommand
{
public:
    VisRenderCallback (Visualization *visualization) :
	vtkCommand (),
	mVisualization (visualization) {}
    static VisRenderCallback *New (Visualization *visualization)
    {
	return new VisRenderCallback (visualization);
    }
    void Delete() { delete this; }
    virtual void Execute (vtkObject *caller, unsigned long eventId,
                          void* arguments)
    {
	mVisualization->updateOnRender ();
    }
private:
    Visualization *mVisualization;
};

class VisStartCallback : public vtkCommand
{
public:
    VisStartCallback (Visualization *visualization) :
	vtkCommand (),
	mVisualization (visualization) {}
    static VisStartCallback *New (Visualization *visualization)
    {
	return new VisStartCallback (visualization);
    }
    void Delete() { delete this; }
    virtual void Execute (vtkObject *caller, unsigned long eventId,
                          void* arguments)
    {
	mVisualization->initInteractorCallbacks ();
	mVisualization->updateOnRender ();

	// Prevent this callback from getting used again.
	mVisualization->getRenderer ()->RemoveObserver (this);
    }
private:
    Visualization *mVisualization;
};

class VisMouseCallback : public vtkCommand
{
public:
    VisMouseCallback (Visualization *visualization) :
	vtkCommand (),
	mVisualization (visualization),
	mMousePointerMoved (false) {}
    static VisMouseCallback *New (Visualization *visualization)
    {
	return new VisMouseCallback (visualization);
    }
    void Delete() { delete this; }
    virtual void Execute (vtkObject *caller, unsigned long eventId,
                          void* arguments)
    {
	vtkRenderWindowInteractor *iren = mVisualization->getRenderWindow ()->
	    GetInteractor ();
	vtkInteractorStyleTrackballCamera *style =
	    (vtkInteractorStyleTrackballCamera *)iren->GetInteractorStyle ();
	switch (eventId)
	{
	case vtkCommand::LeftButtonPressEvent:
	    mMousePointerMoved = false;
	    style->OnLeftButtonDown ();
	    break;
	case vtkCommand::LeftButtonReleaseEvent:
	    if (!mMousePointerMoved)
	    {
		int *pos = iren->GetEventPosition();
		mVisualization->handleClick (pos[0], pos[1]);
	    }
	    style->OnLeftButtonUp ();
	    break;
	case vtkCommand::MouseMoveEvent:
	    mMousePointerMoved = true;
	    style->OnMouseMove ();
	    break;
	}
    }
private:
    Visualization *mVisualization;
    bool mMousePointerMoved;
};

//======= Visualization =======

Visualization::Visualization (Simulation &simulation,
			      VisualizationType visType,
			      QLabel *labelPoint) :
    mSimulation (simulation),
    mVisType (visType),
    mNumDims (mSimulation.getParameters().n),
    mLabelPoint (labelPoint)
{
    mRenderer = vtkRenderer::New();
    mRenWin = vtkRenderWindow::New();
    mRenWin->AddRenderer (mRenderer);

    mColorTable = vtkLookupTable::New ();
    mColorTable->SetHueRange (0.5, 0.5);
    mColorTable->SetSaturationRange (0.0, 0.0);
    mColorTable->SetValueRange (0.0, 1.0);
    mColorTable->SetNumberOfColors (NUM_COLORS);
    mColorTable->Build ();
}

/// Clean up
Visualization::~Visualization ()
{
    mColorTable->Delete ();
    mRenWin->Delete ();
    mRenderer->Delete ();
}

void
Visualization::init ()
{
    VisStartCallback* onStart = VisStartCallback::New (this);
    mRenderer->AddObserver (vtkCommand::StartEvent, onStart);
}

void
Visualization::initInteractorCallbacks ()
{
    VisRenderCallback* onUpdate = VisRenderCallback::New (this);
    mRenWin->GetInteractor ()->AddObserver (vtkCommand::RenderEvent, onUpdate);

    vtkInteractorStyleTrackballCamera *style =
	static_cast<vtkInteractorStyleTrackballCamera *>
	(mRenWin->GetInteractor()->GetInteractorStyle());

    VisMouseCallback* mouseCallback = VisMouseCallback::New (this);

    style->AddObserver (vtkCommand::LeftButtonPressEvent, mouseCallback);
    style->AddObserver (vtkCommand::MouseMoveEvent, mouseCallback);
    style->AddObserver (vtkCommand::LeftButtonReleaseEvent, mouseCallback);
}

//======= TerrainVisualization =======

TerrainVisualization::TerrainVisualization (Simulation &simulation,
                                            QLabel *labelPoint) :
    Visualization (simulation, VisTypeTerrain, labelPoint),
    DotVisualization (simulation, VisTypeTerrain, labelPoint),
    mNumDimsX (mNumDims / 2),
    mNumDimsY (mNumDims - mNumDimsX),
    mNeighborArcsOn (false),
    mSelectedPntGridIndex (0),
    mSelectedPointGroup (NULL)
{
    int nMaxArcs = mNumDims - 2;

    mArcActors = new vtkActor *[nMaxArcs];
    mTerrainActor = vtkActor::New();

    for (int i = 0; i < nMaxArcs; i++)
	mArcActors[i] = vtkActor::New();

    mCaptionActor = vtkCaptionActor2D::New ();
}

TerrainVisualization::~TerrainVisualization ()
{
    //grid2PolyFilter->Delete(); // TODO
    //gridLinesMapper->Delete(); // TODO
    mGridLinesActor->Delete ();

    //terrainMapper->Delete(); // TODO
    mTerrainActor->Delete ();
    //camera->Delete(); // TODO

    int nMaxArcs = mNumDims - 2;
    for (int i = 0; i < nMaxArcs; i++)
	mArcActors[i]->Delete ();

    if (mSelectedPointGroup)
	delete mSelectedPointGroup;

    mTerrainGrid->Delete ();
    mCaptionActor->Delete ();

    delete[] mPointNDimCoords;
}

void
TerrainVisualization::init ()
{
    Visualization::init ();

    // Set up the landscape grid

    mGridResX = 1 << mNumDimsX;
    mGridResY = 1 << mNumDimsY;
    mNumGridPoints = mGridResX * mGridResY;
    int gridRes[3] = {mGridResX, mGridResY, 1};

    printf ("\n");
    printf ("Grid resolution: %u x %u\n", mGridResX, mGridResY);

    mTerrainGrid = vtkStructuredGrid::New();
    mTerrainGrid->SetDimensions (gridRes);

    vtkPoints *points = vtkPoints::New ();
    points->Allocate (mNumGridPoints);
    mPointData =
	static_cast<vtkFloatArray *> (points->GetData ())->GetPointer (0);

    vtkFloatArray *pointsFitness = vtkFloatArray::New();
    pointsFitness->Allocate (mNumGridPoints);
    mPointFitnessData = pointsFitness->GetPointer (0);

    mPointNDimCoords = new unsigned long[mNumGridPoints];

    mScaleFactor = 1.0 / MIN (mGridResX, mGridResY);
    mScaleZFactor = 1.0 / 7;

    initTerrainCoords ();

    mTerrainGrid->SetPoints(points);
    mTerrainGrid->GetPointData()->SetScalars(pointsFitness);

    points->Delete();
    points = NULL;

    // Landscape terrain
    vtkStructuredGridGeometryFilter *grid2PolyFilter =
	vtkStructuredGridGeometryFilter::New ();
    grid2PolyFilter->SetInput (mTerrainGrid);
    grid2PolyFilter->SetExtent (0, mGridResX - 1, 0, mGridResY - 1, 0, 0);

    vtkPolyDataMapper *terrainMapper = vtkPolyDataMapper::New();
    terrainMapper->SetInputConnection(grid2PolyFilter->GetOutputPort());
    terrainMapper->SetScalarRange(0, 1);
    terrainMapper->ScalarVisibilityOn ();
    terrainMapper->SetLookupTable(mColorTable);

    mTerrainActor->SetMapper(terrainMapper);
    mTerrainActor->GetProperty ()->LightingOff ();

    mRenderer->AddActor(mTerrainActor);

    createTerrainGridLines (grid2PolyFilter);

    createSelectedPointDot ();

    initCamera ();

    mRenderer->SetBackground (0.9, 0.9, 0.9);
    mRenWin->PointSmoothingOn (); // Makes dots circular

    createBoundingBox (0, (double)(mGridResX - 1) / mGridResX,
		       0, (double)(mGridResY - 1) / mGridResY,
		       0, mScaleZFactor);
}

void
TerrainVisualization::initCamera ()
{
    vtkCamera *camera = vtkCamera::New ();
    camera->SetPosition (0.5, 0.7, 0.3);
    camera->SetFocalPoint (0.5, 1, 0);

    mRenderer->SetActiveCamera (camera);
    mRenderer->ResetCamera ();
    camera->Dolly (1.3);
    double *clipRange = camera->GetClippingRange();
    camera->SetClippingRange (0.1, clipRange[1]);
}

/// Returns true if a point is picked.
bool
TerrainVisualization::handleClick (int x, int y)
{
    VTK_CREATE_NEW (vtkPointPicker, picker);
    picker->InitializePickList ();
    picker->AddPickList (mTerrainActor);
    picker->PickFromListOn ();

    bool picked = picker->Pick (x, y, 0, mRenderer);
    if (picked)
    {
	vtkIdType pntId = picker->GetPointId ();
	mSelectedPntGridIndex = pntId;

	update ();
    }

    return picked;
}

/// Create grid lines for terrain
void
TerrainVisualization::createTerrainGridLines (vtkStructuredGridGeometryFilter *grid2PolyFilter)
{
    vtkCleanPolyData *cleanFilter = vtkCleanPolyData::New ();
    cleanFilter->SetInputConnection (grid2PolyFilter->GetOutputPort());
    cleanFilter->Update ();

    vtkExtractEdges *gridLines = vtkExtractEdges::New ();
    gridLines->SetInputConnection (cleanFilter->GetOutputPort());

    vtkTubeFilter *tubes = vtkTubeFilter::New ();
    tubes->SetInputConnection (gridLines->GetOutputPort());
    tubes->SetRadius(0.001);
    tubes->SetNumberOfSides(3);

    vtkPolyDataMapper *gridLinesMapper = vtkPolyDataMapper::New ();
    gridLinesMapper->SetInputConnection (tubes->GetOutputPort ());
    gridLinesMapper->SetScalarRange(0, 1);
    gridLinesMapper->ScalarVisibilityOff ();

    mGridLinesActor = vtkActor::New ();
    mGridLinesActor->SetMapper(gridLinesMapper);
    updateGridLineColor ();

    mRenderer->AddActor(mGridLinesActor);
}

/// Create bounding box
void
Visualization::createBoundingBox (double x1, double x2,
				  double y1, double y2,
				  double z1, double z2)
{
    vtkCubeAxesActor *axes = vtkCubeAxesActor::New ();
    axes->SetBounds (x1, x2, y1, y2, z1, z2);
    axes->SetCamera (mRenderer->GetActiveCamera ());
    axes->GetProperty ()->SetColor (0, 0, 0);
    axes->SetFlyModeToStaticEdges ();

    axes->XAxisLabelVisibilityOff ();
    axes->YAxisLabelVisibilityOff ();
    axes->ZAxisLabelVisibilityOff ();

    axes->XAxisTickVisibilityOff ();
    axes->YAxisTickVisibilityOff ();
    axes->ZAxisTickVisibilityOff ();

    mRenderer->AddViewProp (axes);
}

void
TerrainVisualization::toggleNeighborArcs (bool newState)
{
    mNeighborArcsOn = newState;
    if (!newState)
    {
	// Disable arcs
	for (int i = 0; i < mNumArcs; i++)
	    mRenderer->RemoveActor (mArcActors[i]);
	return;
    }

    // Enable arcs
    createNeighborArcs ();

    // Center dot
    createSelectedPointDot ();
}

/// Creates arcs to distant neighbors
void
TerrainVisualization::createNeighborArcs ()
{
    Landscape &landscape = mSimulation.getLandscape ();

    Point pnt (mPointNDimCoords[mSelectedPntGridIndex]);
    unsigned p1r = mSelectedPntGridIndex >> mNumDimsX;              // drop right half
    unsigned p1c = mSelectedPntGridIndex & ((1L << mNumDimsX) - 1) ; // drop left half

    mNumArcs = 0;
    for (int d = 0; d < mNumDims; d++)
    {
	Point pnt2 (pnt, d); // flip ith bit to get a neighbor

	unsigned long p2GridCoords = pointNDimToGridIndex (pnt2.getHash ());
	unsigned p2r = p2GridCoords >> mNumDimsX;              // drop right half
	unsigned p2c = p2GridCoords & ((1 << mNumDimsX) - 1) ; // drop left half

	int distRow = abs ((int)p1r - (int)p2r);
	int distCol = abs ((int)p1c - (int)p2c);

	double p2Fit = landscape.getFitnessValue (pnt2);

	if (distRow == 1 || distCol == 1) // if distance on grid == 1
	    continue; // skip (no arc for grid neighbors)

	double tubeRadius = 0.002;
	float *p1 = &mPointData[3 * mSelectedPntGridIndex];
	float *p2 = &mPointData[3 * p2GridCoords];
	double p1a[3] = {p1[0],
			 p1[1],
			 p1[2] + 2 * tubeRadius *
			 log2 (distRow + distCol - 1) - tubeRadius};
			 //p1[2] + 0.01 * (distRow + distCol) * scaleFactor};
	double inPnt[3] = {(p1[0] + p2[0]) / 2,
			   (p1[1] + p2[1]) / 2,
			   p1[2] + 0.3 * (distRow + distCol) * mScaleFactor};

	vtkPoints* points = vtkPoints::New();
	points->InsertNextPoint (p1a);
	points->InsertNextPoint (inPnt);
	points->InsertNextPoint (p2);

	vtkParametricSpline* spline = vtkParametricSpline::New();
	spline->SetPoints(points);

	vtkParametricFunctionSource* splineSource = vtkParametricFunctionSource::New();
	splineSource->SetUResolution(40); // Set this to how many line segments you
	          //would like in your poly line. More produces a smoother curve.

	splineSource->SetParametricFunction(spline);

	vtkTubeFilter *tube = vtkTubeFilter::New ();
	tube->SetInputConnection(splineSource->GetOutputPort());
	tube->SetRadius(tubeRadius);
	tube->SetNumberOfSides(6);

	vtkPolyDataMapper *arcMapper = vtkPolyDataMapper::New ();
	arcMapper->SetInputConnection (tube->GetOutputPort ());

	mArcActors[mNumArcs]->SetMapper (arcMapper);
	double arcColor = p2Fit;// + 0.12;//(p1Fit + p2Fit) / 2;
	mArcActors[mNumArcs]->GetProperty ()->SetColor (arcColor, arcColor, arcColor);
	mArcActors[mNumArcs]->GetProperty ()->LightingOff ();

	mRenderer->AddActor (mArcActors[mNumArcs]);
	mNumArcs++;
    }
}

void
TerrainVisualization::createSelectedPointDot ()
{
    if (mSelectedPointGroup)
    {
	delete mSelectedPointGroup;
	mRenderer->RemoveActor2D (mCaptionActor);
    }
    unsigned long dataInd = 3 * mSelectedPntGridIndex;
    double pointCoords[3] = {mPointData[dataInd],
                            mPointData[dataInd + 1],
                            mPointData[dataInd + 2] + 0.003};
    double fitness = mPointFitnessData[mSelectedPntGridIndex];

    mSelectedPointGroup =
	new DotGroup (mRenderer, 1, pointCoords,
	              fitness, fitness, fitness, 1.0);

    long pointNDimCoords = mPointNDimCoords[mSelectedPntGridIndex];

    Point pnt (pointNDimCoords);
    string bitStr = pnt.toBitString ();
    QString str;
    str.sprintf ("point: %ld, fitness: %.3f\n%s",
                 pointNDimCoords, fitness, bitStr.c_str ());
    QByteArray ba = str.toLatin1();
    const char *c_str = ba.data();

    //vtkConeSource *coneGlyph = vtkConeSource::New ();
    //coneGlyph->SetResolution (6);

    mCaptionActor->SetCaption (c_str);
    mCaptionActor->GetProperty ()->SetColor (0, 1, 0);
    mCaptionActor->SetAttachmentPoint (pointCoords);
    //mCaptionActor->BorderOff ();
    mCaptionActor->SetPosition (25, 20);
    mCaptionActor->ThreeDimensionalLeaderOff ();
    //mCaptionActor->SetLeaderGlyph (coneGlyph->GetOutput ());
    mCaptionActor->SetWidth (0.30);
    mCaptionActor->SetHeight (0.075);
    //mCaptionActor->SetMaximumLeaderGlyphSize (10);
    //mCaptionActor->SetLeaderGlyphSize (0.05);

    vtkTextProperty *tProp = mCaptionActor->GetCaptionTextProperty ();
    //tProp->SetColor (0, 0, 0);
    //tProp->SetFontFamilyToArial ();
    //tProp->BoldOff ();
    //tProp->SetFontSize(12);
    tProp->ItalicOff ();
    //int shadowOffset[2] = {1, 1};
    //tProp->SetShadowOffset (shadowOffset);
    //tProp->SetColor (1, 1, 1);
    tProp->ShadowOff ();

    mRenderer->AddActor2D (mCaptionActor);
}

void
TerrainVisualization::insertPointDot (const Point &point,
				      DotType dotType,
				      vector<double> *chosenList)
{
    unsigned long pntGridCoords =
	pointNDimToGridIndex (point.getHash ());
    unsigned long gridInd = 3 * pntGridCoords;

    // Prevent point of same memory from jumping as simulation progresses
    srand ((unsigned)pntGridCoords + dotType);
    rand (); // First number is always small, discard it.

    chosenList->push_back (mPointData[gridInd] + 0.0015 * (2.0 * DOUBLE_RAND() - 1));
    chosenList->push_back (mPointData[gridInd + 1] + 0.0015 * (2.0 * DOUBLE_RAND() - 1));
    chosenList->push_back (mPointData[gridInd + 2] + 0.003 + 0.001 * DOUBLE_RAND());
}

void TerrainVisualization::update ()
{
    if (mNeighborArcsOn)
    {
	// Update neighbor arcs
	for (int i = 0; i < mNumArcs; i++)
	    mRenderer->RemoveActor (mArcActors[i]);
	createNeighborArcs ();
    }

    updateTerrainCoords ();

    createSelectedPointDot ();

    mTerrainGrid->Modified ();
    updateGridLineColor ();

    if (mGlobalMaxOn || mLocalMaxOn)
	updatePointDots ();
}

void
TerrainVisualization::initTerrainCoords ()
{
    Landscape &landscape = mSimulation.getLandscape ();

    double totalFitness = 0;
    double minFitness = 2.0;
    double maxFitness = -1;
    unsigned long maxFitPoint = 0;
    unsigned long minFitPoint = 0;

    // Calculate coordinates for all landscape points
    for (long pointCoords = 0; pointCoords < mNumGridPoints;
	pointCoords++)
    {
	unsigned long gridCoords = pointNDimToGridIndex (pointCoords);

	unsigned r = gridCoords >> mNumDimsX;              // drop right half
	unsigned c = gridCoords & ((1L << mNumDimsX) - 1) ; // drop left half

	Point pnt (pointCoords);
	double fitness = landscape.getFitnessValue (pnt);

	unsigned long gridArrayInd  = gridCoords * 3; // for x, y, z
	mPointData[gridArrayInd]     = c * mScaleFactor; // x
	mPointData[gridArrayInd + 1] = r * mScaleFactor; // y
	mPointData[gridArrayInd + 2] = fitness * mScaleZFactor; // z

	mPointFitnessData[gridCoords] = fitness;
	mPointNDimCoords[gridCoords] = pointCoords;

	totalFitness += fitness;
	if (maxFitness < fitness)
	{
	    maxFitness = fitness;
	    maxFitPoint = pointCoords;
	}
	if (minFitness > fitness)
	{
	    minFitness = fitness;
	    minFitPoint = pointCoords;
	}
    }
    mAvgFitness = totalFitness / mNumGridPoints;
    printf ("Fitness min: %.4f (%lu), avg: %.4f, max: %.4f (%lu)\n",
            minFitness, minFitPoint, mAvgFitness, maxFitness, maxFitPoint);

    updateMaxPoints ();
}

void
TerrainVisualization::updateTerrainCoords ()
{
    // Calculate coordinates for all landscape points
    double totalFitness = 0;
    Landscape &landscape = mSimulation.getLandscape ();
    for (long pointCoords = 0; pointCoords < mNumGridPoints;
	pointCoords++)
    {
	Point pnt (pointCoords);
	double fitness = landscape.getFitnessValue (pnt);
	unsigned long gridCoords = pointNDimToGridIndex (pointCoords);
	unsigned long gridArrayInd  = gridCoords * 3; // for x, y, z

	mPointData[gridArrayInd + 2] = fitness * mScaleZFactor; // z
	mPointFitnessData[gridCoords] = fitness;
	totalFitness += fitness;
    }
    mAvgFitness = totalFitness / mNumGridPoints;

    updateMaxPoints ();
}

void
TerrainVisualization::updateMaxPoints ()
{
    double maxFitness = -1;

    for (long pointCoords = 0; pointCoords < mNumGridPoints; pointCoords++)
    {
	Point centerPnt (pointCoords);
	unsigned long gridCoords = pointNDimToGridIndex (pointCoords);
	double centerFitness = mPointFitnessData[gridCoords];

	if (maxFitness < centerFitness)
	{
	    maxFitness = centerFitness;
	    mGlobalMaxPoint = centerPnt;
	}

	int d;
	for (d = 0; d < mNumDims; d++) // for all dimensions
	{
	    Point neighborPnt (centerPnt, d);
	    gridCoords = pointNDimToGridIndex (neighborPnt.getHash ());
	    double neighborFitness = mPointFitnessData[gridCoords];
	    if (neighborFitness > centerFitness)
		break;
	}
	if (d == mNumDims) // if there is no better neighbor
	    mLocalMaxPoints.push_back (centerPnt);
    }
}

/// Adjusts grid line color/opacity using avg. fitness.
void
TerrainVisualization::updateGridLineColor ()
{
    double darkness = (0.5 - mAvgFitness) / 0.5;
    if (darkness < 0)
	darkness = 0;
    // Vary color between (0, 0, 1) and (0.4, 0.8, 1)
    mGridLinesActor->GetProperty ()->
	SetColor (darkness * 0.4, darkness * 0.8, 1);
    // Vary opacity between 0.13 and 0.6
    mGridLinesActor->GetProperty ()->SetOpacity (darkness * 0.6 +
                                                 (1 - darkness) * 0.13);
}

/**
 Converts point's binary representation 2D-grid index
 (i.e. unfolds N-dim to the 2D grid).

    n >= 6 (right half)
    000 -> (0, 0) = 000
    001 -> (0, 1) = 001
    011 -> (0, 2) = 010
    010 -> (0, 3) = 011
    110 -> (0, 4) = 100
    111 -> (0, 5) = 101
    101 -> (0, 6) = 110
    100 -> (0, 7) = 111
    ...
    n >= 8 (right half)
    ...
    1000 -> (0, 8) = 1000
    1001 -> (0, 9) = 1001
    1011 -> (0,10) = 1010
    1010 -> (0,11) = 1011
    1110 -> (0,12) = 1100
    1111 -> (0,13) = 1101
    1101 -> (0,14) = 1110
    1100 -> (0,15) = 1111
*/
unsigned long
TerrainVisualization::pointNDimToGridIndex (unsigned long pointCoords)
{
    unsigned long gridInd = 0;

    unsigned dimOffset = 0;
    unsigned nHalfDims = mNumDims >> 1;
    for (unsigned gridDim = 0; gridDim < 2; gridDim++, dimOffset+= nHalfDims)
    {
	bool flip = false;
	unsigned firstDim = mNumDims - 1 - dimOffset;
	unsigned lastDim = mNumDims - (dimOffset + nHalfDims);
	unsigned long mask = 1L << firstDim;

	for (int d = firstDim; d >= (int)lastDim; d--, mask >>= 1)
	{
	    unsigned long bitVal = pointCoords & mask;
	    if (flip)
		gridInd |= bitVal ^ mask; // use flipped bitVal
	    else
		gridInd |= bitVal;

	    if (bitVal)
		flip = !flip;
	}
    }

    return gridInd;
}

void
TerrainVisualization::pointNDimToGridRowColumn (unsigned long pointCoords,
                                                unsigned &r, unsigned &c)
{
    unsigned long gridCoords = pointNDimToGridIndex (pointCoords);

    r = gridCoords >> mNumDimsX;              // drop right half
    c = gridCoords & ((1L << mNumDimsX) - 1) ; // drop left half
}

#if 0
/** Inverse of pointNDimToGridIndex. That is,
    it converts point's binary representation 2D-grid index
    (i.e. folds the 2D grid into N-dim).
*/
unsigned long
TerrainVisualization::pointGridIndexTONDim (unsigned long pointGridIndex)
{
    return pointGridIndex; // TODO
}
#endif

void
TerrainVisualization::updateOnRender ()
{
    if (mSelectedPointGroup)
	mSelectedPointGroup->update (false);

    postUpdateDots ();
}

void
TerrainVisualization::postUpdateDots ()
{
    for (int i = 0; i < mNumDotGroups; i++)
	if (mDotGroups[i])
	    mDotGroups[i]->update (false);
}
